open Protocol;;

module MySet = Set.Make ((struct type t = int * int;; let compare = compare;; end))
;;

let debug s = print_endline s; flush stdout;;

let money = ref 0;;
let (x, y) = ref 0, ref 0;;
let (width, height) = ref 0, ref 0;;
let bid = ref 6;;

(*plansza, list pol do ktorych chcemy poznac odleglosci i drogi -> odpowiedzi*)
let dijkstra board dest =
  let tab = Array.make_matrix (!width + 2) (!height + 2) (-1, []) in
  tab.(!x).(!y) <- 0, [];
  let act_list = ref [!x, !y] in
  let done_list = ref [] in
  let propagator (x, y) =
    let (dist, route) = tab.(x).(y) in
    let check_field (nx, ny) dir =
      let (d, _) = tab.(nx).(ny) in
      let terrain = board.(nx).(ny) in
      if terrain = Plain && (d == -1 || d > dist + 1) then
        begin
          tab.(nx).(ny) <- dist + 1, (if route = [] then [dir] else route);
          done_list := (nx, ny) :: !done_list
        end;
      if terrain = Coast && (d == -1 || d > dist + 2) then
        begin
          tab.(nx).(ny) <- dist + 2, (if route = [] then [dir] else route);
          done_list := (nx, ny) :: !done_list
				end
    in
    check_field (x + 1, y) E;
    check_field (x - 1, y) W;
    check_field (x, y + 1) N;
    check_field (x, y - 1) S;
    ()
  in
  while !act_list != [] do
    List.iter propagator !act_list; act_list := !done_list; done_list := []
  done;
  List.map (fun (x, y) -> tab.(x).(y)) dest
;;

let big_cost = 1000000000;;

let nearest_of board dest : int * Protocol.dir list =
  let find_nearest (ncost, nroute) (cost, route) =
    if cost < ncost then cost, route else ncost, nroute
  in
  List.fold_left find_nearest (big_cost, []) (dijkstra board dest)
;;

let draw_board board bases =
  let tab = Array.make_matrix !width !height '_' in
  for y = 1 to !height do
    for x = 1 to !width do
      let terrain =
        match board.(x).(y) with
          Wall -> '#'
        | Water -> '~'
        | Plain -> '.'
        | Coast -> ','
      in
      tab.(x - 1).(!height - y) <- terrain
    done
  done;
  let player_drawer uid (x, y) =
    tab.(x - 1).(!height - y) <- char_of_int (uid mod 26 + 65)
  and base_drawer (x, y) = tab.(x - 1).(!height - y) <- '@' in
  Players.for_all player_drawer;
  List.iter base_drawer bases;
  for y = 0 to !height - 1 do
    for x = 0 to !width - 1 do print_char tab.(x).(y) done; print_newline ()
  done;
  print_newline ()
;;

let do_nothing () = Protocol.send_command 1 (Protocol.Pick []);;

let find_reachable board bases =
  let res = dijkstra board bases in
  let combined = List.combine bases res in
  let filtered = List.filter (fun (_, (cost, _)) -> cost <> -1) combined in
  fst (List.split filtered)
;;

(* daje liste wpolrzednych paczek *)
let get_packages_pos packages : (int * int) list = 
	List.map (fun (_, x, y, _) -> x, y) packages;;

let heur_packages board packages =
  let res = dijkstra board (get_packages_pos packages) in
  let combined = List.combine packages res in
  let filtered = List.filter (fun (_, (cost, _)) -> cost <> -1) combined in
  let rank ((_, _, _, w) as package, (cost, _)) =
    (package, (float_of_int w) /. (float_of_int (cost + 1))) in
  let ranked = List.rev_map rank filtered in
  let cmp (_, u) (_, v) =
    if u < v then 1
    else if u > v then -1
    else 0 in
  let sorted = List.sort cmp ranked in
  fst (List.split sorted)
;;

let pickup_packages board capacity packages =
  let heur = heur_packages board packages in
  let rec pick space picked = function (* wybieranie paczek do podniesienia - tu jest miejsce na poprawe :) *)
  | (pid, x, y, weight)::rest ->
      if weight > space then pick space picked rest
      else pick (space - weight) ((pid, x, y, weight)::picked) rest
  | [] -> picked in
  List.rev_map (fun (pid, _, _, _) -> pid) (pick capacity [] heur)
;;

let drop_packages dest =
	let filter (_, x, y, _) = (x, y) = dest in
	let to_drop = List.filter filter !(Players.carried) in	
	List.rev_map (fun (id, _, _, _) -> id) to_drop
;;

let update_bid board opponents =
  if board.(!x-1).(!y) = Water && MySet.mem (!x+1, !y) opponents then bid:= 6
  else if board.(!x+1).(!y) = Water && MySet.mem (!x-1, !y) opponents then bid:= 6
  else if board.(!x).(!y-1) = Water && MySet.mem (!x, !y+1) opponents then bid:= 6
  else if board.(!x).(!y+1) = Water && MySet.mem (!x, !y-1) opponents then bid:= 6
  else bid:= 1
;;

let main () =
  if Array.length Sys.argv != 3 then failwith "argc";
  Protocol.connect Sys.argv.(1) (int_of_string Sys.argv.(2));
  let board, all_bases = Protocol.receive_board () in
  width := Array.length board - 2;
  height := Array.length board.(0) - 2;
  let (my_id, capacity, start_money) = Protocol.receive_config () in
  Players.set_players_pos (Protocol.receive_players ());
  money := start_money;
  let (my_x, my_y) = Players.get_player_pos my_id in
  x := my_x;
  y := my_y; (* updateujemy nasza pozycje *)
  let bases = ref (find_reachable board all_bases) in (* bedziemy sie zajmowac tylko bazami do ktorych mozna dojsc *)
  while true do
    let packages = Protocol.receive_packages () in
    if !Players.carried = [] then begin
      if !bases = [] then do_nothing () (* jesli nic nie niesiemy i nie ma juz baz to nic nie robimy *)
      else begin (* nic nie niesiemy ale sa bazy *)
        let cost, route = nearest_of board !bases in (* zauwazmy ze mamy do czynienia z bazami do ktorych mozemy dojsc (cost<>-1) *)
        if cost = 0 then begin (* stoimy na bazie *)
          if packages = [] then begin
            bases:= List.filter (fun b -> b <> (!x, !y)) !bases; (* jesli baza jest pusta to usuwamy ja z listy baz *)
            do_nothing () (* i nic nie robimy choc w zasadzie moglibysmy zaczac isc do kolejnej ale to zbyt skomplikowane *)
          end else
            let picked = pickup_packages board capacity packages in (* sa paczki w bazie *)
            Protocol.send_command 1 (Protocol.Pick picked) (* probujemy podniesc - czy nam sie uda odpowie serwer *)
        end else begin (* musimy dojsc do bazy *)
        	match List.rev route with
						| dir::rest -> Protocol.send_command !bid (Protocol.Move dir) (* probujemy isc - czy sie uda odpowie serwer *)
						| [] -> failwith "route empty, and not there"
        end
      end
    end else begin (* niesiemy jakies paczki *)
      let cost, route = nearest_of board (get_packages_pos !Players.carried) in (* zauwazmy ze mamy do czynienia z paczkami do ktorych przeznaczenia mozemy dojsc, bo tylko takie podnosilismy *)
      if cost = 0 then (* doszlismy do miejsca przeznaczenia paczki(czek) *)
        Protocol.send_command 1 (Protocol.Drop (drop_packages (!x, !y))) (* probujemy upuscic odpowiednie paczki - czy sie uda odpowie serwer *)
      else begin (* musimy dojsc do najblizszego miejsca przeznaczenia *)
        match List.rev route with
					| dir::rest -> Protocol.send_command !bid (Protocol.Move dir) (* probujemy isc - czy sie uda odpowie serwer *)
					| [] -> failwith "route empty, and not there"
      end
    end;
    let changes, alive = Protocol.receive_changes ()in
    Players.update_players packages my_id changes;
    let rec find_opponents opponents = function
    | uid::rest -> find_opponents (MySet.add (Players.get_player_pos uid) opponents) rest
    | [] -> opponents in
    update_bid board (find_opponents MySet.empty alive);
    (*draw_board board !bases;*)
    let (my_x, my_y) = Players.get_player_pos my_id in
    x := my_x;
    y := my_y(*; (* updateujemy nasza pozycje *)
		let j= ref 1 in
		for i =1 to 3000000 do
			j:= !j+1
		done*)
  done;
  ()
;;
  (*Protocol.disconnect in_chann*)

main ();;

