open Protocol;;

module IntMap = Map.Make ((struct type t = int;; let compare = compare;; end))
;;

module IntSet = Set.Make ((struct type t = int;; let compare = compare;; end))
;;

let players : (int * int) IntMap.t ref = ref IntMap.empty;;

let get_player_pos uid = IntMap.find uid !players;;

let carried = ref [];;	

let set_players_pos desc =
  let iterator (id, x, y) = players := IntMap.add id (x, y) !players in
  List.iter iterator desc
;;

let set_player_pos uid (x, y) = players := IntMap.add uid (x, y) !players;;
		
let move_player uid (dx, dy) =
  let (x, y) = get_player_pos uid in set_player_pos uid (x + dx, y + dy)
;;
		
let update_players packages my_id changes =
	let pick_set = ref IntSet.empty 
	and drop_set = ref IntSet.empty in
  let iterator (uid, action) =
    match action with
      AMove N -> move_player uid (0, 1)
    | AMove S -> move_player uid (0, -1)
    | AMove W -> move_player uid (-1, 0)
    | AMove E -> move_player uid (1, 0)
    | ANew (x, y) -> set_player_pos uid (x, y)
    | APick pid ->
        if my_id = uid then
					pick_set := IntSet.add pid !pick_set
    | ADrop pid ->
        if my_id = uid then
					drop_set := IntSet.add pid !drop_set
  in
  List.iter iterator changes; 
	if (not (IntSet.is_empty !drop_set)) then
		let filter (id, _, _, _) = not (IntSet.mem id !drop_set) in
		carried := List.filter filter !carried
	else ();
	if (not (IntSet.is_empty !pick_set)) then
		let iterator ((id, x, y, sth) as pack) = 
			if IntSet.mem id !pick_set then carried := pack :: !carried else ()
		in
			List.iter iterator packages
	else ()
;;

let for_all funct = IntMap.iter funct !players;;
