type cell = Plain | Water | Wall | Coast;;
type dir = N | W | S | E;;

type command = Move of dir | Pick of int list | Drop of int list;;

type action = AMove of dir | ADrop of int | APick of int | ANew of int * int;;

let (in_chann, out_chann) = ref stdin, ref stdout;;

let dir2char =
  function
    N -> 'N'
  | S -> 'S'
  | W -> 'W'
  | E -> 'E'
;;

let send_command bid comm =
  output_string !out_chann (string_of_int bid);
  output_char !out_chann ' ';
  begin match comm with
    Move N -> output_string !out_chann "Move N"
  | Move S -> output_string !out_chann "Move S"
  | Move E -> output_string !out_chann "Move E"
  | Move W -> output_string !out_chann "Move W"
  | Pick pids ->
      output_string !out_chann "Pick";
      List.iter
        (fun pid ->
           output_char !out_chann ' ';
           output_string !out_chann (string_of_int pid)
				)
        pids
  | Drop pids ->
      output_string !out_chann "Drop";
      List.iter
        (fun pid ->
           output_char !out_chann ' ';
           output_string !out_chann (string_of_int pid))
        pids
  end;
  output_char !out_chann '\n';
  flush !out_chann
;;

let receive_board () =
  let cell_of_char =
    function
      '.' -> Plain
    | '~' -> Water
    | '#' -> Wall
    | '@' -> Plain
    | _ -> failwith "cell_of_char zasyfilo"
  in
  match
    Genlex.make_lexer [] (Stream.of_string (input_line !in_chann)) with
    parser
    [< 'Genlex.Int width; 'Genlex.Int height >] ->
      let board = Array.make_matrix (width + 2) (height + 2) Wall
      and bases = ref [] in
      for y = 1 to height do
        let row = input_line !in_chann in
        for x = 1 to width do
          board.(x).(y) <- cell_of_char row.[x - 1];
          if row.[x - 1] = '@' then bases := (x, y) :: !bases
        done
      done;
      for y = 1 to height do
        for x = 1 to width do
					if board.(x).(y) = Plain && 
					(	board.(x-1).(y)=Water
					|| board.(x+1).(y)=Water
					|| board.(x).(y-1)=Water
					|| board.(x).(y+1)=Water
					) then board.(x).(y) <- Coast
				else ()
        done
      done;
      board, !bases
;;

let receive_config () =
  match
    Genlex.make_lexer [] (Stream.of_string (input_line !in_chann)) with
    parser
  [< 'Genlex.Int uid; 'Genlex.Int mcc; 'Genlex.Int money >] -> uid, mcc, money
;;


let receive_players () =
  let rec receive_aux =
    parser
      [< 'Genlex.Kwd "#"; 'Genlex.Int uid; 'Genlex.Kwd "X"; 'Genlex.Int x;
         'Genlex.Kwd "Y"; 'Genlex.Int y; s >] ->
        (uid, x, y) :: receive_aux s
    | [< >] -> []
  in
  let str = Stream.of_string (input_line !in_chann) in
  receive_aux (Genlex.make_lexer ["#"; "X"; "Y"] str)
;;


let receive_packages () =
  let rec receive_aux =
    parser
      [< 'Genlex.Int pid; 'Genlex.Int dest_x; 'Genlex.Int dest_y;
         'Genlex.Int weight; s >] ->
        (pid, dest_x, dest_y, weight) :: receive_aux s
    | [< >] -> []
  in
  let str = Stream.of_string (input_line !in_chann) in
  receive_aux (Genlex.make_lexer [] str)
;;


let receive_changes () =
  let alive = ref [] in
  let rec receive_aux uid =
    parser
      [< 'Genlex.Kwd "#"; 'Genlex.Int new_uid; s >] -> alive:= new_uid::!alive; receive_aux new_uid s
    | [< 'Genlex.Kwd "N"; s >] -> (uid, AMove N) :: receive_aux uid s
    | [< 'Genlex.Kwd "S"; s >] -> (uid, AMove S) :: receive_aux uid s
    | [< 'Genlex.Kwd "E"; s >] -> (uid, AMove E) :: receive_aux uid s
    | [< 'Genlex.Kwd "W"; s >] -> (uid, AMove W) :: receive_aux uid s
    | [< 'Genlex.Kwd "P"; 'Genlex.Int pid; s >] ->
        (uid, APick pid) :: receive_aux uid s
    | [< 'Genlex.Kwd "D"; 'Genlex.Int pid; s >] ->
        (uid, ADrop pid) :: receive_aux uid s
    | [< 'Genlex.Kwd "X"; 'Genlex.Int x; 'Genlex.Kwd "Y"; 'Genlex.Int y;
         s >] ->
        (uid, ANew (x, y)) :: receive_aux uid s
    | [< >] -> []
  in
  let new_line = input_line !in_chann in
  let str = Stream.of_string new_line in
  let lexer =
    Genlex.make_lexer ["#"; "N"; "S"; "E"; "W"; "P"; "D"; "X"; "Y"] str
  in
  receive_aux 0 lexer, !alive
;;


let connect host port =
  let inet_addr = (Unix.gethostbyname host).Unix.h_addr_list.(0) in
  let (in_ch, out_ch) =
    Unix.open_connection (Unix.ADDR_INET (inet_addr, port))
  in
  in_chann := in_ch;
  out_chann := out_ch;
  output_string !out_chann "Player\n";
  flush !out_chann
;;

let disconnect () =
  Unix.shutdown_connection !in_chann; in_chann := stdin; out_chann := stdout
;;
